// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
extern "C" {
  #include <archive.h>
  #include <archive_entry.h>
}
#include "dawnlibs.hpp"
#include "assert/assert.hpp"
#include "util/memory.hpp"

#define ASSET_LOADER_BUFFER_SIZE 32768

#if !defined(DAWN_ASSET_LOCATION)
  #error Asset Archive Location has not been defined.
#endif

/**
 * Method invoked by the libarchive internals to read bytes from the archive
 * file pointer.
 * 
 * @param archive Archive requesting the read. 
 * @param data Data pointer passed to the archive.
 * @param buffer Pointer to where the buffer pointer should be stored.
 * @return Count of bytes read.
 */
ssize_t assetLoaderArchiveRead(
  struct archive *archive,
  void *data,
  const void **buffer
);

/**
 * Method invoked by the libarchive internals to seek the archive file pointer.
 * 
 * @param archive Archive requesting the seek.
 * @param data Data pointer passed to the archive.
 * @param offset Offset to seek to.
 * @param whence Whence to seek from.
 * @return The new offset.
 */
int64_t assetLoaderArchiveSeek(
  struct archive *archive,
  void *data,
  int64_t offset,
  int32_t whence
);

/**
 * Method invoked by the libarchive internals to open the archive file pointer.
 * 
 * @param archive Archive requesting the open.
 * @param data Data pointer passed to the archive.
 * @return 0 if success, otherwise for failure.
 */
int32_t assetLoaderArchiveOpen(struct archive *a, void *data);

/**
 * Method invoked by the libarchive internals to close the archive file pointer.
 * 
 * @param archive Archive requesting the close.
 * @param data Data pointer passed to the archive.
 * @return 0 if success, otherwise for failure.
 */
int32_t assetLoaderArchiveClose(struct archive *a, void *data);

namespace Dawn {
  class AssetLoader {
    private:
      std::string fileName;
      struct archive *assetArchive = nullptr;
      struct archive_entry *assetArchiveEntry = nullptr;
      size_t position;

    public:
      uint8_t buffer[ASSET_LOADER_BUFFER_SIZE];
      FILE *assetArchiveFile = nullptr;

      /**
       * Create a new asset loader. Asset Loaders can be used to load data from
       * a file in a myriad of ways.
       * 
       * @param fileName File name of the asset that is to be loaded.
       */
      AssetLoader(std::string fileName);

      /**
       * Platform-centric method to open a file buffer to an asset.
       * 
       * @return 0 if success, otherwise for failure.
       */
      void open();

      /**
       * Closes the previously ppened asset.
       * @return 0 if successful, otherwise false.
       */
      int32_t close();

      /**
       * Read bytes from buffer.
       * @param buffer Pointer to a ubyte array to buffer data into.
       * @param size Length of the data buffer (How many bytes to read).
       * @return The count of bytes read.
       */
      size_t read(uint8_t *buffer, size_t size);

      /**
       * Get the size of the asset.
       * @return The size of the asset in bytes.
       */
      size_t getSize();

      /**
       * Skips the read head forward to a given position.
       * 
       * @param n Count of bytes to progress the read head by.
       * @return Count of bytes progressed.
       */
      size_t skip(size_t n);

      /**
       * Rewind the read head to the beginning of the file.
       */
      void rewind();

      /**
       * Sets the absolute position of the read head within the buffer of the
       * file.
       * 
       * @param absolutePosition Absolute position to set the read head to.
       */
      size_t setPosition(size_t absolutePosition);

      /**
       * Returns the current position of the read head.
       * 
       * @return The current read head position.
       */
      size_t getPosition();

      /**
       * Run a callback for each byte within the asset. The callback will
       * receive each byte individually.
       * 
       * @tparam T Type of instance to run callback against.
       * @param instance Instance of the object to run the callback against.
       * @param callback Callback method on the class to run the callback for.
       * @return The count of bytes read.
       */
      template<class T>
      size_t loadBufferedCallback(T *instance, bool_t (T::*callback)(uint8_t n)) {
        uint8_t buffer[1024];
        size_t read, length;
        int32_t i;
        bool_t result;

        assertNotNull(instance);
        assertNotNull(callback);

        // Open the buffer.
        this->open();

        // Reset length size
        length = 0;

        // Buffer from input
        while((read = this->read(buffer, 1024)) != 0) {
          for(i = 0; i < read; i++) {
            result = ((*instance).*(callback))(buffer[i]);
            if(!result) {
              length += i;
              break;
            }
          }
          if(!result) break;
          length += read;
        }

        // Close the buffer
        this->close();

        return length;
      }

      /**
       * Cleanup the asset loader.
       */
      virtual ~AssetLoader();
  };
}