// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "LocaleManager.hpp"
#include "game/DawnGame.hpp"


using namespace Dawn;

LocaleManager::LocaleManager(DawnGame *game) {
  this->game = game;
  this->locale.language = "en";
}

void LocaleManager::init() {
  // Begin loading
  if(this->currentlyLoadedAsset == nullptr && this->loadingAsset == nullptr) {
    this->loadingAsset = this->game->assetManager.getAndLoad<LanguageAsset>("language_" + this->locale.language);
    if(this->loadingAsset->loaded) {
      this->onLanguageLoaded();
    } else {
      this->loadingAsset->eventLoaded.addListener(this, &LocaleManager::onLanguageLoaded);
    }
  }
}

void LocaleManager::setLocale(struct Locale locale) {
  auto oldLocale = this->getLocale();

  this->locale = locale;
  this->eventLocaleChanged.invoke();

  // Did the language change?
  if(this->locale.language != oldLocale.language) {
    // Yes, switch languages.

    // Are we already loading something? Stop it.
    if(this->loadingAsset != nullptr) {
      this->loadingAsset->eventLoaded.removeListener(this, &LocaleManager::onLanguageLoaded);
      this->game->assetManager.queueUnload(this->loadingAsset);
      this->loadingAsset = nullptr;
    }

    this->loadingAsset = this->game->assetManager.getAndLoad<LanguageAsset>("language_" + this->locale.language);
    if(this->loadingAsset->loaded) {
      this->onLanguageLoaded();
    } else {
      this->loadingAsset->eventLoaded.addListener(this, &LocaleManager::onLanguageLoaded);
    }
  }
}

struct Locale LocaleManager::getLocale() {
  return this->locale;
}

std::string LocaleManager::getString(std::string key) {
  assertNotNull(this->currentlyLoadedAsset, "LocaleManager::getString: Currently loaded asset cannot be null");
  return this->currentlyLoadedAsset->getValue(key);
}


void LocaleManager::onLanguageLoaded() {
  // Unload the previously loaded language
  if(this->currentlyLoadedAsset != nullptr) {
    this->currentlyLoadedAsset->eventLoaded.removeListener(this, &LocaleManager::onLanguageLoaded);
    this->game->assetManager.queueUnload(this->currentlyLoadedAsset);
  }

  // Swap loaded languages
  this->currentlyLoadedAsset = this->loadingAsset;
  this->loadingAsset = nullptr;

  // Notify.
  this->eventLanguageUpdated.invoke();
}