// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "memory.hpp"
#include "assert/assert.hpp"

uint64_t dawnAllocatedItemCount = -1;

void memoryInit() {
  dawnAllocatedItemCount = 0;
}

void memoryDispose() {
  assertTrue(dawnAllocatedItemCount <= 0, "memoryDispose: !!!Leaked memory detected (not enough free)!!!");
  assertTrue(dawnAllocatedItemCount >= 0, "memoryDispose: !!!Leaked memory detected (too much free)!!!");
  dawnAllocatedItemCount = -1;
}

void * memoryAllocate(const size_t size) {
  assertTrue(dawnAllocatedItemCount >= 0, "memoryAllocate: Either a memory leak or memoryInit was not called.");
  assertTrue(size >= 0, "memoryAllocate: size must be greater than 0 or equal to.");
  if(size == 0) {
    // Technically this is an implementation specific case, I think it should
    // basically be undefined behaviour tbh.
    return memoryCallMalloc(size);
  }
  dawnAllocatedItemCount++;
  auto x = (void *)memoryCallMalloc(size);
  assertNotNull(x, "memoryAllocate: Failed to allocate memory");
  return x;
}

void * memoryAllocateEmpty(const size_t count, const size_t size) {
  assertTrue(dawnAllocatedItemCount >= 0, "memoryAllocateEmpty: Either a memory leak or memoryInit was not called.");
  assertTrue(count >= 0, "memoryAllocateEmpty: count must be greater than or equal to 0");
  assertTrue(size >= 0, "memoryAllocateEmpty: size must be greater than or equal to 0");
  if(count == 0 || size == 0) {
    // Technically this is an implementation specific case, I think it should
    // basically be undefined behaviour tbh.
    return memoryCallCalloc(count, size);
  }
  dawnAllocatedItemCount++;
  auto x = (void*)memoryCallCalloc(count, size);
  assertNotNull(x, "memoryAllocateEmpty: Failed to allocate memory");
  return x;
}

void memoryFree(void *pointer) {
  assertTrue(dawnAllocatedItemCount > 0, "memoryFree: Either a memory leak or memoryInit was not called.");
  if(pointer == nullptr || pointer == NULL) {
    // Technically this is an implementation specific case, I think it should
    // basically be undefined behaviour tbh.
    return memoryCallFree(pointer);
  }
  dawnAllocatedItemCount--;
  memoryCallFree(pointer);
}

void memoryCopy(void *source, void *destination, size_t size) {
  assertTrue(dawnAllocatedItemCount >= 0, "memoryAllocate: Either a memory leak or memoryInit was not called.");
  assertNotNull(source, "memoryCopy: source must not be null");
  assertNotNull(destination, "memoryCopy: destination must not be null");
  assertTrue(destination != source, "memoryCopy: destination must not be source");
  assertTrue(size > 0, "memoryCopy: size must be greater than 0");
  memcpy(destination, source, size);
}

int32_t memoryCompare(const void *left, const void *right, const size_t size) {
  assertTrue(dawnAllocatedItemCount >= 0, "memoryAllocate: Either a memory leak or memoryInit was not called.");
  assertTrue(left != right, "memoryCompare: left must not be right");
  assertTrue(size > 0, "memoryCompare: size must be greater than 0");
  return memcmp(left, right, size);
}

void memorySet(void *dest, uint8_t data, size_t length) {
  assertTrue(dawnAllocatedItemCount >= 0, "memoryAllocate: Either a memory leak or memoryInit was not called.");
  assertNotNull(dest, "memorySet: dest must not be null");
  assertTrue(length > 0, "memorySet: length must be greater than 0");
  memset(dest, data, length);
}

size_t memoryReallocate(void **pointer, size_t currentSize, size_t newSize) {
  // Create the new buffer
  void *newBuffer = memoryAllocate(newSize);
  memoryCopy(*pointer, newBuffer, currentSize);
  memoryFree(*pointer);
  *pointer = newBuffer;
  return newSize;
}

void * operator new(size_t size) {
  return memoryAllocate(size);
}

void operator delete(void *pointer) {
  memoryFree(pointer);
}